from functools import lru_cache
from typing import Tuple, Dict, List, Optional
from pathlib import Path

import dovado_rtl.vivado_interaction as vivado
import dovado_rtl.report_parsing as report
from dovado_rtl.frame_handling import (
    HdlBoxFrameHandler,
    TclFrameHandler,
)
from dovado_rtl.config import Configuration
from dovado_rtl.simple_types import IsIncremental, DesignValue, Metric
from dovado_rtl.enums import StopStep
from dovado_rtl.cli_utility import ask_utilization_metrics
from dovado_rtl.report_parsing import get_available_indices
from dovado_rtl.abstract_classes import (
    AbstractDesignPointEvaluator,
    AbstractEstimator,
    AbstractSourceParser,
)


class DesignPointEvaluator(AbstractDesignPointEvaluator):
    def __init__(
        self,
        config: Configuration,
        parsed_source: AbstractSourceParser,
        hdl_handler: HdlBoxFrameHandler,
        tcl_handler: TclFrameHandler,
        target_clock: float,
        is_incremental: IsIncremental,
        stop_step: StopStep,
        free_parameters: List[str],
        int_metrics: Optional[List[int]],
    ):
        self.__hdl_handler: HdlBoxFrameHandler = hdl_handler
        self.__stop_step: StopStep = stop_step
        self.__parsed_file = parsed_source
        self.__target_clock: float = target_clock
        self.__is_incremental: IsIncremental = is_incremental
        self.__is_first_evaluation: bool = True
        self.__free_parameters: List[str] = free_parameters
        self.__config: Configuration = config
        self.__tcl_handler: TclFrameHandler = tcl_handler
        self.__metrics: Optional[List[Metric]] = None
        self.__int_metrics: Optional[List[int]] = int_metrics
        self.__estimator: Optional[AbstractEstimator] = None

        self.__out_file = Path(config.get_config("WORK_DIR")).joinpath(
            "space_exploration.csv"
        )
        self.__out_file_mapping: Dict[str, Metric] = {}
        self.__out_file.open("w").close()
        self.__is_file_header_to_write = True

    @lru_cache()
    def evaluate(self, design_point: Tuple[int, ...]) -> Optional[DesignValue]:

        self.__parsed_file.write_parameter_values(
            self.__hdl_handler,
            dict(zip(self.__free_parameters, design_point)),
        )

        vivado_out, success = vivado.source(
            str(self.__config.get_config("WORK_DIR"))
            + str(self.__config.get_config(self.__stop_step.name))
        )

        print(vivado_out)
        if self.__is_first_evaluation:
            if not success:
                return None
            if not self.__metrics:
                self.__metrics = ask_utilization_metrics(
                    get_available_indices(
                        str(self.__config.get_config("WORK_DIR"))
                        + str(
                            self.__config.get_config(
                                self.__stop_step.name + "_UTILISATION"
                            )
                        )
                    ),
                    Path(str(self.__config.get_config("CUSTOM_METRICS_DIR"))),
                    self.__int_metrics,
                )
            self.__is_first_evaluation = False
        if not success:
            met: Dict[Metric, float] = {}
            for i in self.__metrics:
                if i.is_frequency or i.custom_metric:
                    met[i] = float(0)
                else:
                    met[i] = float(100)
            design_value = DesignValue(met)
        else:
            design_value = DesignValue(
                value={
                    i: report.get_utilisation(
                        str(self.__config.get_config("WORK_DIR"))
                        + (
                            str(
                                self.__config.get_config(
                                    self.__stop_step.name + "_UTILISATION"
                                )
                            )
                        ),
                        i.utilisation[0],
                        i.utilisation[1],
                    )
                    if not i.is_frequency
                    else -self.get_max_frequency(
                        report.get_wns(
                            str(self.__config.get_config("WORK_DIR"))
                            + (
                                str(
                                    self.__config.get_config(
                                        self.__stop_step.name + "_TIMING"
                                    )
                                )
                            )
                        )
                    )
                    for i in self.__metrics
                    if not i.custom_metric
                },
            )
            for i in self.__metrics:
                if i.custom_metric:
                    design_value.value[i] = self.compute_custom_metric(
                        i, design_point
                    )
        self.__write_csv(design_point, design_value)
        return design_value

    def get_max_frequency(self, wns: float) -> float:

        return 1000 / (1 / (1 / 1000 * self.__target_clock) - wns)

    def compute_custom_metric(
        self, metric: Metric, design_point: Tuple[int, ...]
    ):
        if not metric.custom_metric:
            raise Exception(
                "Trying to execute metric '" + str(metric) + "' as custom"
            )

        all_metrics = get_available_indices(
            str(self.__config.get_config("WORK_DIR"))
            + str(
                self.__config.get_config(
                    self.__stop_step.name + "_UTILISATION"
                )
            )
        )

        metrics_value = {}
        for section in all_metrics.keys():
            for m in all_metrics[section]:
                metrics_value[m] = report.get_utilisation(
                    str(self.__config.get_config("WORK_DIR"))
                    + (
                        str(
                            self.__config.get_config(
                                self.__stop_step.name + "_UTILISATION"
                            )
                        )
                    ),
                    section,
                    m,
                )
        metrics_value["frequency"] = self.get_max_frequency(
            report.get_wns(
                str(self.__config.get_config("WORK_DIR"))
                + (
                    str(
                        self.__config.get_config(
                            self.__stop_step.name + "_TIMING"
                        )
                    )
                )
            )
        )

        params_value = {
            fp: dp for fp, dp in zip(self.__free_parameters, design_point)
        }

        return metric.custom_metric[1](**{**metrics_value, **params_value})

    def set_metrics(self, metrics: List[Metric]):
        if self.__metrics:
            raise Exception(
                "Metrics already set with value: "
                + str(self.__metrics)
                + " resetting forbidden. Trying to reset with value "
                + str(metrics)
            )
        self.__metrics = metrics

    def __write_csv(
        self, design_point: Tuple[int, ...], design_value: DesignValue
    ):
        if self.__is_file_header_to_write and self.__metrics:
            header = ",".join(self.__free_parameters)
            for m in self.__metrics:
                header += ","
                if m.utilisation:
                    header += m.utilisation[1]
                    self.__out_file_mapping[m.utilisation[1]] = m
                elif m.is_frequency:
                    header += "Frequency"
                    self.__out_file_mapping["Frequency"] = m
                elif m.custom_metric:
                    header += m.custom_metric[0]
                    self.__out_file_mapping[m.custom_metric[0]] = m

            header += "\n"
            self.__is_file_header_to_write = False
            self.__out_file.open(mode="a+").writelines(header)
        with open(self.__out_file) as file:
            header = file.readline()
        line = ",".join([str(d) for d in design_point])
        header = header[:-1].split(",")[len(design_point) :]
        for metric in header:
            line += ","
            line += str(design_value.value[self.__out_file_mapping[metric]])
        line += "\n"
        self.__out_file.open(mode="a+").writelines(line)

    def set_estimator(self, estimator: AbstractEstimator) -> None:
        self.__estimator = estimator

    def get_metrics(self):
        return self.__metrics

    def get_parsed_src(self):
        return self.__parsed_file



from dovado_rtl.antlr.hdl_representation import (
    Entity,
    ParameterType,
    ParameterTypeEnum,
    Port,
    Parameter,
    PortDirection,
    PortDirectionEnum,
    PortType,
    PortTypeEnum,
)
from itertools import chain
from typing import List, Tuple, Iterator, Union
from dovado_rtl.antlr.generated.SystemVerilog.SystemVerilogParserVisitor import (
    SystemVerilogParserVisitor,
)
from dovado_rtl.antlr.generated.SystemVerilog.SystemVerilogParser import (
    SystemVerilogParser,
)


class SysVerilogHDLEntityVisitor(SystemVerilogParserVisitor):
    def __init__(self):
        self.entities: List[Entity] = []
        self.top_level = None

    def visitSource_text(
        self, ctx: SystemVerilogParser.Source_textContext
    ) -> None:
        for d in ctx.description():
            self.visitDescription(d)

    def visitDescription(
        self, ctx: SystemVerilogParser.DescriptionContext
    ) -> None:
        if ctx.module_declaration():
            self.visitModule_declaration(ctx.module_declaration())
        else:
            pass

    def visitModule_declaration(
        self, ctx: SystemVerilogParser.Module_declarationContext
    ) -> None:
        if ctx.module_nonansi_header():
            self.visitModule_nonansi_header(ctx.module_nonansi_header())
        elif ctx.module_ansi_header():
            self.visitModule_ansi_header(ctx.module_ansi_header())
        else:
            self.entities.append(
                Entity(self.visitModule_identifier(ctx.module_identifier()[0]))
            )

    def visitModule_nonansi_header(
        self, ctx: SystemVerilogParser.Module_nonansi_headerContext
    ):
        self.entities.append(
            Entity(self.visitModule_identifier(ctx.module_identifier()))
        )
        if ctx.parameter_port_list():
            self.visitParameter_port_list(ctx.parameter_port_list())
        print("Non-Ansi ports found, skipping...")
        # self.visitList_of_ports(ctx.list_of_ports())

    def visitList_of_ports(
        self, ctx: SystemVerilogParser.List_of_portsContext
    ):
        for i in ctx.port():
            self.visitPort(i)

    def visitList_of_port_declarations(
        self, ctx: SystemVerilogParser.List_of_port_declarationsContext
    ):
        for i in ctx.ansi_port_declaration():
            self.visitAnsi_port_declaration(i)

    def visitAnsi_port_declaration(
        self, ctx: SystemVerilogParser.Ansi_port_declarationContext
    ):
        direction = None
        if ctx.net_port_header():
            direction = self.visitNet_port_header(ctx.net_port_header())
        if ctx.variable_port_header():
            direction = self.visitVariable_port_header(
                ctx.variable_port_header()
            )
        if ctx.port_direction():
            direction = self.visitPort_direction()
        if direction:
            ident = self.visitPort_identifier(ctx.port_identifier())

            self.entities[-1].add_port(Port(ident, direction, None))

    def visitNet_port_header(
        self, ctx: SystemVerilogParser.Net_port_headerContext
    ):
        if ctx.port_direction():
            return self.visitPort_direction(ctx.port_direction())
        return None

    def visitVariable_port_header(
        self, ctx: SystemVerilogParser.Variable_port_headerContext
    ):
        if ctx.port_direction():
            return self.visitPort_direction(ctx.port_direction())
        return None

    def visitPort_direction(
        self, ctx: SystemVerilogParser.Port_directionContext
    ):
        if ctx.KINPUT():
            return PortDirection(PortDirectionEnum.INPUT, descriptor="input")
        if ctx.KOUTPUT():
            return PortDirection(PortDirectionEnum.OUTPUT, descriptor="output")
        if ctx.KINOUT():
            return PortDirection(PortDirectionEnum.OTHER, descriptor="inout")
        return None

    def visitModule_ansi_header(
        self, ctx: SystemVerilogParser.Module_ansi_headerContext
    ):
        self.entities.append(
            Entity(self.visitModule_identifier(ctx.module_identifier()))
        )
        if ctx.parameter_port_list():
            self.visitParameter_port_list(ctx.parameter_port_list())
        if ctx.list_of_port_declarations():
            self.visitList_of_port_declarations(
                ctx.list_of_port_declarations()
            )

    def visitParameter_port_list(
        self, ctx: SystemVerilogParser.Parameter_port_listContext
    ):
        for i in ctx.parameter_port_declaration():
            self.visitParameter_port_declaration(i)
        if ctx.list_of_param_assignments():
            self.visitList_of_param_assignments(
                ctx.list_of_param_assignments()
            )

    def visitList_of_param_assignments(
        self, ctx: SystemVerilogParser.List_of_param_assignmentsContext
    ):
        for i in ctx.param_assignment():
            yield self.visitParam_assignment(i)

    def visitParam_assignment(
        self, ctx: SystemVerilogParser.Param_assignmentContext
    ):

        return self.visitParameter_identifier(ctx.parameter_identifier())

    def visitParameter_identifier(
        self, ctx: SystemVerilogParser.Parameter_identifierContext
    ):
        return self.visitIdentifier(ctx.identifier())

    def visitParameter_port_declaration(
        self, ctx: SystemVerilogParser.Parameter_port_declarationContext
    ):
        if ctx.parameter_declaration():
            self.visitParameter_declaration(ctx.parameter_declaration())

    def visitList_of_port_declarations(
        self, ctx: SystemVerilogParser.List_of_port_declarationsContext
    ) -> None:
        for p in ctx.ansi_port_declaration():
            self.visitAnsi_port_declaration(p)

    def visitList_of_port_declarations_comma(self, ctx) -> None:
        self.visitAttr_port_declaration(ctx.attr_port_declaration())
        self.visitComma_attr_port_declaration_star(
            ctx.comma_attr_port_declaration_star()
        )

    def visitComma_attr_port_declaration_star(self, ctx) -> None:
        if not ctx.comma_attr_port_declaration():
            return
        for a in ctx.comma_attr_port_declaration():
            self.visitComma_attr_port_declaration(a)

    def visitComma_attr_port_declaration(self, ctx) -> None:
        self.visitAttr_port_declaration(ctx.attr_port_declaration())

    def visitList_of_port_declarations_semicolon(self, ctx) -> None:
        self.visitAttr_port_declaration_semicolon_plus(
            ctx.attr_port_declaration_semicolon_plus()
        )

    def visitAttr_port_declaration_semicolon_plus(self, ctx) -> None:
        for a in ctx.attr_port_declaration_semicolon():
            self.visitAttr_port_declaration_semicolon(a)

    def visitAttr_port_declaration_semicolon(self, ctx) -> None:
        self.visitAttr_port_declaration(ctx.attr_port_declaration())

    def visitAttr_port_declaration(self, ctx) -> None:
        self.visitPort_declaration(ctx.port_declaration())

    def visitPort_declaration(
        self, ctx: SystemVerilogParser.Port_declarationContext
    ) -> None:
        if ctx.inout_declaration():
            self.visitInout_declaration(ctx.inout_declaration())
        if ctx.input_declaration():
            self.visitInput_declaration(ctx.input_declaration())
        if ctx.output_declaration():
            self.visitOutput_declaration(ctx.output_declaration())
        if ctx.ref_declaration():
            print("Ref ports not supported, skipping")

    def visitInout_declaration(
        self, ctx: SystemVerilogParser.Inout_declarationContext
    ) -> None:
        for ident in self.visitList_of_port_identifiers(
            ctx.list_of_port_identifiers()
        ):
            self.entities[-1].add_port(
                Port(
                    ident,
                    PortDirection(PortDirectionEnum.OTHER, "inout"),
                    None,
                )
            )

    def visitInput_declaration(
        self, ctx: SystemVerilogParser.Input_declarationContext
    ) -> None:
        for ident in self.visitList_of_port_identifiers(
            ctx.list_of_port_identifiers()
        ):
            self.entities[-1].add_port(
                Port(
                    ident,
                    PortDirection(PortDirectionEnum.INPUT, "input"),
                    None,
                )
            )

    def visitOutput_declaration(
        self, ctx: SystemVerilogParser.Output_declarationContext
    ) -> None:
        for ident in self.visitList_of_port_identifiers(
            ctx.list_of_port_identifiers()
        ):
            self.entities[-1].add_port(
                Port(
                    ident,
                    PortDirection(PortDirectionEnum.OUTPUT, "output"),
                    None,
                )
            )

    def visitList_of_port_identifiers(
        self, ctx: SystemVerilogParser.List_of_port_identifiersContext
    ) -> None:
        for i in self.visitPort_identifier(ctx.port_identifier()):
            yield self.visitPort_identifier(i)

    def visitPort_identifier(
        self, ctx: SystemVerilogParser.Port_identifierContext
    ) -> str:
        return self.visitIdentifier(ctx.identifier())

    def visitList_of_interface_parameters(self, ctx) -> None:
        if ctx.list_of_parameter_declarations():
            self.visitList_of_parameter_declarations(
                ctx.list_of_parameter_declarations()
            )
        if ctx.list_of_parameter_descriptions():
            self.visitList_of_parameter_descriptions(
                ctx.list_of_parameter_descriptions()
            )

    def visitList_of_parameter_declarations(self, ctx) -> None:
        self.visitParameter_declaration(ctx.parameter_declaration())
        self.visitComma_parameter_declaration_star(
            ctx.comma_parameter_declaration_star()
        )

    def visitComma_parameter_declaration_star(self, ctx) -> None:
        if not ctx.comma_parameter_declaration():
            return
        for p in ctx.comma_parameter_declaration():
            self.visitComma_parameter_declaration(p)

    def visitComma_parameter_declaration(self, ctx) -> None:
        self.visitParameter_declaration(ctx.parameter_declaration())

    def visitParameter_declaration(
        self, ctx: SystemVerilogParser.Parameter_declarationContext
    ) -> None:
        param_type = self.visitData_type_or_implicit(
            ctx.data_type_or_implicit()
        )
        if param_type:
            for param in self.visitList_of_param_assignments(
                ctx.list_of_param_assignments()
            ):
                self.entities[-1].add_parameter(
                    Parameter(str(param), param_type)
                )

    def visitData_type(self, ctx: SystemVerilogParser.Data_typeContext):
        if ctx.integer_atom_type():
            return self.visitInteger_atom_type(ctx.integer_atom_type())
        else:
            return None

    def visitInteger_atom_type(
        self, ctx: SystemVerilogParser.Integer_atom_typeContext
    ):
        if ctx.KBYTE():
            return ParameterType(ParameterTypeEnum.INTEGER, "byte")
        if ctx.KSHORTINT():
            return ParameterType(ParameterTypeEnum.INTEGER, "shortint")
        if ctx.KINT():
            return ParameterType(ParameterTypeEnum.INTEGER, "int")
        if ctx.KLONGINT():
            return ParameterType(ParameterTypeEnum.INTEGER, "longint")
        if ctx.KINTEGER():
            return ParameterType(ParameterTypeEnum.INTEGER, "integer")
        if ctx.KTIME():
            return ParameterType(ParameterTypeEnum.INTEGER, "time")

    def visitParam_description(self, ctx) -> None:
        if ctx.param_declaration():
            for v in self.visitParam_declaration(ctx.param_declaration()):
                self.entities[-1].add_parameter(Parameter(v[0], v[1]))

        for d in chain(
            self.visitLogic_declaration(ctx.logic_declaration())
            if ctx.logic_declaration()
            else iter(()),
            self.visitInteger_declaration(ctx.integer_declaration())
            if ctx.integer_declaration()
            else iter(()),
            self.visitInt_declaration(ctx.int_declaration())
            if ctx.int_declaration()
            else iter(()),
            self.visitTime_declaration(ctx.time_declaration())
            if ctx.time_declaration()
            else iter(()),
        ):
            self.entities[-1].add_parameter(Parameter(d[0], d[1]))
        if ctx.usertype_variable_declaration():
            print("UserType parameters are not supported, skipping")
        else:
            # non numeric or non synthesizable types skipped
            pass

    def visitParam_declaration(self, ctx) -> List[Tuple[str, ParameterType]]:
        out = []
        if ctx.Signed():
            for v in self.visitList_of_hierarchical_variable_descriptions(
                ctx.list_of_hierarchical_variable_descriptions()
            ):
                out.append(
                    (v, ParameterType(ParameterTypeEnum.INTEGER, "Signed"))
                )
        if ctx.Unsigned():
            for v in self.visitList_of_hierarchical_variable_descriptions(
                ctx.list_of_hierarchical_variable_descriptions()
            ):
                out.append(
                    (v, ParameterType(ParameterTypeEnum.INTEGER, "Unsigned"))
                )
        else:
            for v in self.visitList_of_hierarchical_variable_descriptions(
                ctx.list_of_hierarchical_variable_descriptions()
            ):
                out.append((v, ParameterType(ParameterTypeEnum.INTEGER, "")))
        return out

    def visitBits_declaration(
        self, ctx
    ) -> Iterator[Tuple[str, ParameterType]]:
        for v in self.visitList_of_variable_descriptions(
            ctx.list_of_variable_descriptions()
        ):
            b_type = self.visitBits_type(ctx.bits_type())
            b_type += "signed" if ctx.Signed() else ""
            b_type += "unsigned" if ctx.Unsigned() else ""
            yield v, ParameterType(ParameterTypeEnum.INTEGER, b_type)

    def visitBits_type(self, ctx) -> str:
        if ctx.Bit():
            return ctx.Bit().getText()
        else:
            return ctx.Byte().getText()

    def visitNet_declaration(
        self, ctx: SystemVerilogParser.Net_declarationContext
    ) -> Iterator[Tuple[str, PortType]]:
        for v in self.visitList_of_variable_descriptions(
            ctx.list_of_variable_descriptions()
        ):
            net: PortType = self.visitNet_type(ctx.net_type())
            p_type = net.descriptor
            p_type += "vectored" if ctx.Vectored() else ""
            p_type += "scalared" if ctx.Scalared() else ""
            p_type += "signed" if ctx.Signed() else ""
            p_type += "unsigned" if ctx.Unsigned() else ""
            yield v, PortType(net.type, p_type)

    def visitNet_type(
        self, ctx: SystemVerilogParser.Net_typeContext
    ) -> PortType:
        if ctx.Supply0():
            return PortType(PortTypeEnum.VECTOR, ctx.Supply0().getText())
        if ctx.Supply1():
            return PortType(PortTypeEnum.VECTOR, ctx.Supply1().getText())
        if ctx.Tri():
            return PortType(PortTypeEnum.VECTOR, ctx.Tri().getText())
        if ctx.Tri_and():
            return PortType(PortTypeEnum.VECTOR, ctx.Tri_and().getText())
        if ctx.Tri_or():
            return PortType(PortTypeEnum.VECTOR, ctx.Tri_or().getText())
        if ctx.Tri_reg():
            return PortType(PortTypeEnum.VECTOR, ctx.Tri_reg().getText())
        if ctx.Tri0():
            return PortType(PortTypeEnum.VECTOR, ctx.Tri0().getText())
        if ctx.Tri1():
            return PortType(PortTypeEnum.VECTOR, ctx.Tri1().getText())
        if ctx.Uwire():
            return PortType(PortTypeEnum.VECTOR, ctx.Uwire().getText())
        if ctx.Wire():
            return PortType(PortTypeEnum.VECTOR, ctx.Wire().getText())
        if ctx.Wand():
            return PortType(PortTypeEnum.VECTOR, ctx.Wand().getText())
        if ctx.Wor():
            return PortType(PortTypeEnum.VECTOR, ctx.Wor().getText())
        return PortType(PortTypeEnum.VECTOR, ctx.NONE().getText())

    def visitLogic_declaration(
        self, ctx
    ) -> Union[
        Iterator[Tuple[str, ParameterType]], Iterator[Tuple[str, PortType]]
    ]:
        for v in self.visitList_of_variable_descriptions(
            ctx.list_of_variable_descriptions()
        ):
            if isinstance(
                ctx.parentCtx.parentCtx.parentCtx,
                SystemVerilogParser.Port_declarationContext,
            ):
                yield v, PortType(PortTypeEnum.SCALAR, "Logic")

            else:
                yield v, ParameterType(ParameterTypeEnum.INTEGER, "Logic")

    def visitInteger_declaration(
        self, ctx
    ) -> Union[
        Iterator[Tuple[str, ParameterType]], Iterator[Tuple[str, PortType]]
    ]:
        for v in self.visitList_of_variable_descriptions(
            ctx.list_of_variable_descriptions()
        ):
            p_type = "automatic" if ctx.Automatic() else ""
            p_type += "integer"
            p_type += "signed" if ctx.Signed() else ""
            p_type += "unsigned" if ctx.Unsigned() else ""
            if isinstance(
                ctx.parentCtx.parentCtx.parentCtx,
                SystemVerilogParser.Port_declarationContext,
            ):
                yield v, PortType(PortTypeEnum.SCALAR, p_type)
            else:
                yield v, ParameterType(ParameterTypeEnum.INTEGER, p_type)

    def visitInt_declaration(
        self, ctx
    ) -> Union[
        Iterator[Tuple[str, ParameterType]], Iterator[Tuple[str, PortType]]
    ]:
        for v in self.visitList_of_variable_descriptions(
            ctx.list_of_variable_descriptions()
        ):

            p_type = "automatic" if ctx.Automatic() else ""
            p_type += "static" if ctx.Static() else ""
            p_type += "const" if ctx.Const() else ""
            p_type += "integer"
            p_type += "signed" if ctx.Signed() else ""
            p_type += "unsigned" if ctx.Unsigned() else ""
            if isinstance(
                ctx.parentCtx.parentCtx.parentCtx,
                SystemVerilogParser.Port_declarationContext,
            ):
                yield v, PortType(PortTypeEnum.SCALAR, p_type)
            else:
                yield v, ParameterType(ParameterTypeEnum.INTEGER, p_type)

    def visitTime_declaration(
        self, ctx
    ) -> Union[
        Iterator[Tuple[str, ParameterType]], Iterator[Tuple[str, PortType]]
    ]:

        for v in self.visitList_of_variable_descriptions(
            ctx.list_of_variable_descriptions()
        ):
            if isinstance(
                ctx.parentCtx.parentCtx.parentCtx,
                SystemVerilogParser.Port_declarationContext,
            ):
                yield v, PortType(PortTypeEnum.SCALAR, "Time")
            else:
                yield v, ParameterType(ParameterTypeEnum.INTEGER, "Time")

    def visitList_of_hierarchical_variable_descriptions(
        self, ctx
    ) -> List[str]:
        vds = self.visitHierarchical_variable_description(
            ctx.hierarchical_variable_description()
        )

        vds.extend(
            self.visitComma_hierarchical_variable_description_star(
                ctx.comma_hierarchical_variable_description_star()
            )
        )
        return vds

    def visitComma_hierarchical_variable_description_star(
        self, ctx
    ) -> List[str]:
        out = []
        for h in ctx.comma_hierarchical_variable_description():
            out.extend(self.visitComma_hierarchical_variable_description(h))
        return out

    def visitComma_hierarchical_variable_description(self, ctx) -> List[str]:
        return self.visitHierarchical_variable_description(
            ctx.hierarchical_variable_description()
        )

    def visitHierarchical_variable_description(self, ctx) -> List[str]:
        return self.visitHierarchical_variable_identifier(
            ctx.hierarchical_variable_identifier()
        )

    def visitHierarchical_variable_identifier(
        self, ctx: SystemVerilogParser.Hierarchical_variable_identifierContext
    ) -> List[str]:
        return self.visitHierarchical_identifier(ctx.hierarchical_identifier())

    def visitHierarchical_identifier(
        self, ctx: SystemVerilogParser.Hierarchical_identifierContext
    ) -> List[str]:
        out = [
            self.visitHierarchical_identifier_branch_item(
                ctx.hierarchical_identifier_branch_item()
            )
        ]
        out.extend(
            self.visitDot_hierarchical_identifier_branch_item_star(
                ctx.dot_hierarchical_identifier_branch_item_star()
            )
        )
        return out

    def visitDot_hierarchical_identifier_branch_item_star(
        self, ctx
    ) -> List[str]:
        out = []
        for h in ctx.dot_hierarchical_identifier_branch_item():
            out.append(self.visitDot_hierarchical_identifier_branch_item(h))
        return out

    def visitDot_hierarchical_identifier_branch_item(self, ctx) -> str:
        return self.visitHierarchical_identifier_branch_item(
            ctx.hierarchical_identifier_branch_item()
        )

    def visitHierarchical_identifier_branch_item(self, ctx):
        return self.visitIdentifier(ctx.identifier())

    def visitList_of_variable_descriptions(
        self, ctx: SystemVerilogParser
    ) -> List[str]:
        vds = [self.visitVariable_description(ctx.variable_description())]
        vds.extend(
            self.visitComma_variable_description_star(
                ctx.comma_variable_description_star()
            )
        )
        return vds

    def visitComma_variable_description_star(
        self, ctx: SystemVerilogParser
    ) -> List[str]:
        vds = []
        for v in ctx.comma_variable_description():
            vds.append(self.visitComma_variable_description(v))
        return vds

    def visitComma_variable_description(self, ctx: SystemVerilogParser) -> str:
        return self.visitVariable_description(ctx.variable_description())

    def visitVariable_description(self, ctx: SystemVerilogParser) -> str:
        return self.visitVariable_identifier(ctx.variable_identifier())

    def visitVariable_identifier(
        self, ctx: SystemVerilogParser.Variable_identifierContext
    ) -> str:
        return self.visitIdentifier(ctx.identifier())

    def visitModule_identifier(
        self, ctx: SystemVerilogParser.Module_identifierContext
    ) -> str:
        return self.visitIdentifier(ctx.identifier())

    def visitIdentifier(
        self, ctx: SystemVerilogParser.IdentifierContext
    ) -> str:
        return ctx.SIMPLE_IDENTIFIER().getText()



from dovado_rtl.antlr.hdl_representation import (
    ParameterType,
    ParameterTypeEnum,
    PortDirection,
    Entity,
    Port,
    Parameter,
    PortDirectionEnum,
    PortType,
    PortTypeEnum,
)
from dovado_rtl.antlr.generated.Verilog2001.VerilogParserVisitor import (
    VerilogParserVisitor,
)
from dovado_rtl.antlr.generated.Verilog2001.VerilogParser import VerilogParser
from typing import List, Optional


class Verilog2001EntityVisitor(VerilogParserVisitor):
    def __init__(self):
        self.entities: List[Entity] = []
        self.top_level = None

    def visitSource_text(self, ctx: VerilogParser.Source_textContext) -> None:
        for d in ctx.description():
            self.visitDescription(d)

    def visitDescription(self, ctx: VerilogParser.DescriptionContext) -> None:
        self.visitModule_declaration(ctx.module_declaration())

    def visitModule_declaration(
        self, ctx: VerilogParser.Module_declarationContext
    ) -> None:
        self.entities.append(
            Entity(self.visitModule_identifier(ctx.module_identifier()))
        )
        if ctx.module_parameter_port_list():
            self.visitModule_parameter_port_list(
                ctx.module_parameter_port_list()
            )
        if ctx.list_of_ports():
            self.visitList_of_ports(ctx.list_of_ports())
        if ctx.list_of_port_declarations():
            self.visitList_of_port_declarations(
                ctx.list_of_port_declarations()
            )
        for m in ctx.module_item():
            self.visitModule_item(m)
        for nm in ctx.non_port_module_item():
            self.visitNon_port_module_item(nm)

    def visitNon_port_module_item(
        self, ctx: VerilogParser.Non_port_module_itemContext
    ):
        if ctx.module_or_generate_item():
            self.visitModule_or_generate_item(ctx.module_or_generate_item())
        if ctx.parameter_declaration():
            self.visitParameter_declaration(ctx.parameter_declaration())
        if ctx.specify_block():
            self.visitSpecify_block(ctx.specify_block())
        if ctx.specparam_declaration():
            self.visitSpecparam_declaration(ctx.specparam_declaration())
        pass

    def visitModule_item(self, ctx: VerilogParser.Module_itemContext) -> None:
        if ctx.port_declaration():
            self.visitPort_declaration(ctx.port_declaration())
        if ctx.non_port_module_item():
            self.visitNon_port_module_item(ctx.non_port_module_item)

    def visitSpecify_block(
        self, ctx: VerilogParser.Specify_blockContext
    ) -> None:
        for i in ctx.specify_item():
            self.visitSpecify_item(i)

    def visitSpecify_item(
        self, ctx: VerilogParser.Specify_itemContext
    ) -> None:
        if ctx.specparam_declaration():
            self.visitSpecparam_declaration(ctx.specparam_declaration())
        else:
            pass

    def visitSpecparam_declaration(
        self, ctx: VerilogParser.Specparam_declarationContext
    ) -> None:
        for p in self.visitList_of_specparam_assignments(
            ctx.list_of_specparam_assignments()
        ):
            self.entities[-1].add_parameter(
                Parameter(
                    p, ParameterType(ParameterTypeEnum.INTEGER, "specparam")
                )
            )

    def visitList_of_specparam_assignments(
        self, ctx: VerilogParser.List_of_specparam_assignmentsContext
    ) -> List[str]:
        specparams = []
        for s in ctx.specparam_assignment():
            specparams.append(self.visitSpecparam_assignment(s))
        return specparams

    def visitSpecparam_assignment(
        self, ctx: VerilogParser.Specparam_assignmentContext
    ) -> Optional[str]:
        if ctx.specparam_identifier():
            return self.visitSpecparam_identifier(ctx.specparam_identifier())
        else:
            pass

    def visitSpecparam_identifier(
        self, ctx: VerilogParser.Specparam_identifierContext
    ) -> str:
        return self.visitIdentifier(ctx.identifier())

    def visitModule_or_generate_item(
        self, ctx: VerilogParser.Module_or_generate_itemContext
    ):
        if ctx.parameter_override():
            self.visitParameter_override(ctx.parameter_override())
        else:
            pass

    def visitParameter_override(
        self, ctx: VerilogParser.Parameter_overrideContext
    ) -> None:
        for p in self.visitList_of_parameter_assignments(
            ctx.list_of_param_assignments()
        ):
            self.entities[-1].add_parameter(Parameter(p, None))

    def visitLocal_parameter_declaration(
        self, ctx: VerilogParser.Local_parameter_declarationContext
    ):
        if ctx.range_():
            for p in self.visitList_of_param_assignments(
                ctx.list_of_param_assignments()
            ):
                self.entities[-1].add_parameter(Parameter(p, None))
        else:
            for p in self.visitList_of_param_assignments(
                ctx.list_of_param_assignments()
            ):
                self.entities[-1].add_parameter(
                    Parameter(p, ctx.getText().split()[1])
                )

    def visitList_of_port_declarations(
        self, ctx: VerilogParser.List_of_port_declarationsContext
    ) -> None:
        for pd in ctx.port_declaration():
            self.visitPort_declaration(pd)

    def visitPort_declaration(
        self, ctx: VerilogParser.Port_declarationContext
    ) -> None:
        if ctx.attribute_instance():
            pass
        if ctx.inout_declaration():
            self.visitInout_declaration(ctx.inout_declaration())
        if ctx.input_declaration():
            self.visitInput_declaration(ctx.input_declaration())
        if ctx.output_declaration():
            self.visitOutput_declaration(ctx.output_declaration())

    def visitInout_declaration(
        self, ctx: VerilogParser.Inout_declarationContext
    ) -> None:
        for p in ctx.list_of_port_identifiers():
            self.entities[-1].add_port(
                Port(
                    p,
                    PortDirection(PortDirectionEnum.OTHER, "inout"),
                    PortType(PortTypeEnum.VECTOR, "wire/reg")
                    if not ctx.net_type()
                    else self.visitNet_type(ctx.net_type()),
                )
            )

    def visitInput_declaration(
        self, ctx: VerilogParser.Input_declarationContext
    ) -> None:
        for p in self.visitList_of_port_identifiers(
            ctx.list_of_port_identifiers()
        ):
            self.entities[-1].add_port(
                Port(
                    p,
                    PortDirection(PortDirectionEnum.INPUT, "input"),
                    PortType(PortTypeEnum.VECTOR, "wire/reg")
                    if not ctx.net_type()
                    else self.visitNet_type(ctx.net_type()),
                )
            )

    def visitOutput_declaration(
        self, ctx: VerilogParser.Output_declarationContext
    ) -> None:
        if ctx.list_of_port_identifiers():
            for p in self.visitList_of_port_identifiers(
                ctx.list_of_port_identifiers()
            ):
                self.entities[-1].add_port(
                    Port(
                        p,
                        PortDirection(PortDirectionEnum.OUTPUT, "output"),
                        (
                            PortType(PortTypeEnum.VECTOR, "wire/reg")
                            if not ctx.net_type()
                            else self.visitNet_type(ctx.net_type())
                        )
                        if not ctx.output_variable_type()
                        else ctx.output_variable_type(),
                    )
                )

        if ctx.list_of_variable_port_identifiers():
            for vp in self.visitList_of_variable_port_identifiers(
                ctx.list_of_variable_port_identifiers()
            ):
                Port(
                    vp,
                    PortDirection(PortDirectionEnum.OUTPUT, "output"),
                    (
                        PortType(PortTypeEnum.VECTOR, "wire/reg")
                        if not ctx.net_type()
                        else self.visitNet_type(ctx.net_type())
                    )
                    if not ctx.output_variable_type()
                    else ctx.output_variable_type(),
                )

    def visitList_of_variable_port_identifiers(
        self, ctx: VerilogParser.List_of_variable_port_identifiersContext
    ) -> List[str]:
        ids = []
        for p in ctx.port_identifier():
            ids.append(self.visitPort_identifier(p))

        if ctx.constant_expression():
            pass
        return ids

    def visitList_of_port_identifiers(
        self, ctx: VerilogParser.List_of_port_identifiersContext
    ) -> List[str]:
        ids = []
        for pi in ctx.port_identifier():
            ids.append(self.visitPort_identifier(pi))
        return ids

    def visitNet_type(self, ctx: VerilogParser.Net_typeContext) -> PortType:
        return PortType(PortTypeEnum.VECTOR, "")

    def visitOutput_variable_type(
        self, ctx: VerilogParser.Output_variable_typeContext
    ) -> PortType:
        return PortType(PortTypeEnum.VECTOR, "")

    def visitList_of_ports(
        self, ctx: VerilogParser.List_of_portsContext
    ) -> None:
        for p in ctx.port():
            self.visitPort(p)

    def visitPort(self, ctx: VerilogParser.PortContext) -> None:
        if ctx.port_identifier():
            self.entities[-1].add_port(
                Port(
                    self.visitPort_identifier(ctx.port_identifier()),
                    None,
                    None,
                )
            )
        else:
            for p in self.visitPort_expression(ctx.port_expression()):
                self.entities[-1].add_port(Port(p, None, None))

    def visitPort_identifier(
        self, ctx: VerilogParser.Port_identifierContext
    ) -> str:
        return self.visitIdentifier(ctx.identifier())

    def visitPort_expression(
        self, ctx: VerilogParser.Port_expressionContext
    ) -> List[str]:
        refs = []
        for pr in ctx.port_reference():
            refs.append(self.visitPort_reference(pr))
        return refs

    def visitPort_reference(
        self, ctx: VerilogParser.Port_referenceContext
    ) -> str:
        if ctx.constant_expression():
            pass
        if ctx.range_expression():
            pass

        return self.visitPort_identifier(ctx.port_identifier())

    def visitModule_parameter_port_list(
        self, ctx: VerilogParser.Module_parameter_port_listContext
    ) -> None:
        for p in ctx.parameter_declaration():
            self.visitParameter_declaration_(p)

    def visitParameter_declaration(
        self, ctx: VerilogParser.Parameter_declarationContext
    ) -> None:
        self.visitList_of_param_assignments(ctx.list_of_param_assignments())

    def visitParameter_declaration_(self, ctx) -> None:
        for pa in self.visitList_of_param_assignments(
            ctx.list_of_param_assignments()
        ):
            if "integer" in ctx.getText()[9 : 9 + len("integer")]:
                self.entities[-1].add_parameter(
                    Parameter(
                        pa,
                        ParameterType(ParameterTypeEnum.INTEGER, "integer"),
                    )
                )
            if "time" in ctx.getText()[9 : 9 + len("time")]:
                self.entities[-1].add_parameter(
                    Parameter(
                        pa, ParameterType(ParameterTypeEnum.INTEGER, "time"),
                    )
                )
            else:
                self.entities[-1].add_parameter(
                    Parameter(
                        pa,
                        ParameterType(ParameterTypeEnum.INTEGER, "integer"),
                    )
                )

                pass

    def visitList_of_param_assignments(
        self, ctx: VerilogParser.List_of_param_assignmentsContext
    ) -> List[str]:
        params = []
        for p in ctx.param_assignment():
            params.append(self.visitParam_assignment(p))
        return params

    def visitParam_assignment(
        self, ctx: VerilogParser.Param_assignmentContext
    ) -> str:
        return self.visitParameter_identifier(ctx.parameter_identifier())

    def visitParameter_identifier(
        self, ctx: VerilogParser.Parameter_identifierContext
    ) -> str:
        return self.visitIdentifier(ctx.identifier())

    def visitModule_identifier(
        self, ctx: VerilogParser.Module_identifierContext
    ) -> str:
        return self.visitIdentifier(ctx.identifier())

    def visitIdentifier(self, ctx: VerilogParser.IdentifierContext) -> str:
        if ctx.SIMPLE_IDENTIFIER():
            return ctx.SIMPLE_IDENTIFIER().getText()
        else:
            return ctx.ESCAPED_IDENTIFIER().getText()



# Generated from VerilogParser.g4 by ANTLR 4.9.3
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .VerilogParser import VerilogParser
else:
    from VerilogParser import VerilogParser

# This class defines a complete generic visitor for a parse tree produced by VerilogParser.

class VerilogParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by VerilogParser#display_tasks.
    def visitDisplay_tasks(self, ctx:VerilogParser.Display_tasksContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#display_task_name.
    def visitDisplay_task_name(self, ctx:VerilogParser.Display_task_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_arguments.
    def visitList_of_arguments(self, ctx:VerilogParser.List_of_argumentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#argument.
    def visitArgument(self, ctx:VerilogParser.ArgumentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#strobe_tasks.
    def visitStrobe_tasks(self, ctx:VerilogParser.Strobe_tasksContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#strobe_task_name.
    def visitStrobe_task_name(self, ctx:VerilogParser.Strobe_task_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#monitor_tasks.
    def visitMonitor_tasks(self, ctx:VerilogParser.Monitor_tasksContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#monitor_task_name.
    def visitMonitor_task_name(self, ctx:VerilogParser.Monitor_task_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#file_open_function.
    def visitFile_open_function(self, ctx:VerilogParser.File_open_functionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#file_close_task.
    def visitFile_close_task(self, ctx:VerilogParser.File_close_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#multi_channel_descriptor.
    def visitMulti_channel_descriptor(self, ctx:VerilogParser.Multi_channel_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#fd.
    def visitFd(self, ctx:VerilogParser.FdContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#file_name.
    def visitFile_name(self, ctx:VerilogParser.File_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#type_.
    def visitType_(self, ctx:VerilogParser.Type_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#file_output_tasks.
    def visitFile_output_tasks(self, ctx:VerilogParser.File_output_tasksContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#file_output_task_name.
    def visitFile_output_task_name(self, ctx:VerilogParser.File_output_task_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#load_memory_tasks.
    def visitLoad_memory_tasks(self, ctx:VerilogParser.Load_memory_tasksContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#memory_name.
    def visitMemory_name(self, ctx:VerilogParser.Memory_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#start_addr.
    def visitStart_addr(self, ctx:VerilogParser.Start_addrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#finish_addr.
    def visitFinish_addr(self, ctx:VerilogParser.Finish_addrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#filename.
    def visitFilename(self, ctx:VerilogParser.FilenameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#finish_task.
    def visitFinish_task(self, ctx:VerilogParser.Finish_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#finish_number.
    def visitFinish_number(self, ctx:VerilogParser.Finish_numberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#stop_task.
    def visitStop_task(self, ctx:VerilogParser.Stop_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#time_function.
    def visitTime_function(self, ctx:VerilogParser.Time_functionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#stime_function.
    def visitStime_function(self, ctx:VerilogParser.Stime_functionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#realtime_function.
    def visitRealtime_function(self, ctx:VerilogParser.Realtime_functionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#conversion_functions.
    def visitConversion_functions(self, ctx:VerilogParser.Conversion_functionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#conversion_function_name.
    def visitConversion_function_name(self, ctx:VerilogParser.Conversion_function_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#constant_argument.
    def visitConstant_argument(self, ctx:VerilogParser.Constant_argumentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#random_function.
    def visitRandom_function(self, ctx:VerilogParser.Random_functionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#seed.
    def visitSeed(self, ctx:VerilogParser.SeedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dist_functions.
    def visitDist_functions(self, ctx:VerilogParser.Dist_functionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#start_.
    def visitStart_(self, ctx:VerilogParser.Start_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#end.
    def visitEnd(self, ctx:VerilogParser.EndContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#mean.
    def visitMean(self, ctx:VerilogParser.MeanContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#standard_deviation.
    def visitStandard_deviation(self, ctx:VerilogParser.Standard_deviationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#degree_of_freedom.
    def visitDegree_of_freedom(self, ctx:VerilogParser.Degree_of_freedomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#k_stage.
    def visitK_stage(self, ctx:VerilogParser.K_stageContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#math_functions.
    def visitMath_functions(self, ctx:VerilogParser.Math_functionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#integer_math_functions.
    def visitInteger_math_functions(self, ctx:VerilogParser.Integer_math_functionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#real_math_functions.
    def visitReal_math_functions(self, ctx:VerilogParser.Real_math_functionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#single_argument_real_math_function_name.
    def visitSingle_argument_real_math_function_name(self, ctx:VerilogParser.Single_argument_real_math_function_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#double_argument_real_math_function_name.
    def visitDouble_argument_real_math_function_name(self, ctx:VerilogParser.Double_argument_real_math_function_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumpfile_task.
    def visitDumpfile_task(self, ctx:VerilogParser.Dumpfile_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumpvars_task.
    def visitDumpvars_task(self, ctx:VerilogParser.Dumpvars_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_modules_or_variables.
    def visitList_of_modules_or_variables(self, ctx:VerilogParser.List_of_modules_or_variablesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_or_variable.
    def visitModule_or_variable(self, ctx:VerilogParser.Module_or_variableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#levels.
    def visitLevels(self, ctx:VerilogParser.LevelsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumpoff_task.
    def visitDumpoff_task(self, ctx:VerilogParser.Dumpoff_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumpon_task.
    def visitDumpon_task(self, ctx:VerilogParser.Dumpon_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumpall_task.
    def visitDumpall_task(self, ctx:VerilogParser.Dumpall_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumplimit_task.
    def visitDumplimit_task(self, ctx:VerilogParser.Dumplimit_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#file_size.
    def visitFile_size(self, ctx:VerilogParser.File_sizeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumpflush_task.
    def visitDumpflush_task(self, ctx:VerilogParser.Dumpflush_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumpports_task.
    def visitDumpports_task(self, ctx:VerilogParser.Dumpports_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#scope_list.
    def visitScope_list(self, ctx:VerilogParser.Scope_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#file_pathname.
    def visitFile_pathname(self, ctx:VerilogParser.File_pathnameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumpportsoff_task.
    def visitDumpportsoff_task(self, ctx:VerilogParser.Dumpportsoff_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumpportson_task.
    def visitDumpportson_task(self, ctx:VerilogParser.Dumpportson_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumpportsall_task.
    def visitDumpportsall_task(self, ctx:VerilogParser.Dumpportsall_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumpportslimit_task.
    def visitDumpportslimit_task(self, ctx:VerilogParser.Dumpportslimit_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dumpportsflush_task.
    def visitDumpportsflush_task(self, ctx:VerilogParser.Dumpportsflush_taskContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#library_text.
    def visitLibrary_text(self, ctx:VerilogParser.Library_textContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#library_description.
    def visitLibrary_description(self, ctx:VerilogParser.Library_descriptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#library_declaration.
    def visitLibrary_declaration(self, ctx:VerilogParser.Library_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#include_statement.
    def visitInclude_statement(self, ctx:VerilogParser.Include_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#source_text.
    def visitSource_text(self, ctx:VerilogParser.Source_textContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#description.
    def visitDescription(self, ctx:VerilogParser.DescriptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_declaration.
    def visitModule_declaration(self, ctx:VerilogParser.Module_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_keyword.
    def visitModule_keyword(self, ctx:VerilogParser.Module_keywordContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_parameter_port_list.
    def visitModule_parameter_port_list(self, ctx:VerilogParser.Module_parameter_port_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_ports.
    def visitList_of_ports(self, ctx:VerilogParser.List_of_portsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_port_declarations.
    def visitList_of_port_declarations(self, ctx:VerilogParser.List_of_port_declarationsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#port.
    def visitPort(self, ctx:VerilogParser.PortContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#port_expression.
    def visitPort_expression(self, ctx:VerilogParser.Port_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#port_reference.
    def visitPort_reference(self, ctx:VerilogParser.Port_referenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#port_declaration.
    def visitPort_declaration(self, ctx:VerilogParser.Port_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_item.
    def visitModule_item(self, ctx:VerilogParser.Module_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_or_generate_item.
    def visitModule_or_generate_item(self, ctx:VerilogParser.Module_or_generate_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_or_generate_item_declaration.
    def visitModule_or_generate_item_declaration(self, ctx:VerilogParser.Module_or_generate_item_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#non_port_module_item.
    def visitNon_port_module_item(self, ctx:VerilogParser.Non_port_module_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#parameter_override.
    def visitParameter_override(self, ctx:VerilogParser.Parameter_overrideContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#config_declaration.
    def visitConfig_declaration(self, ctx:VerilogParser.Config_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#design_statement.
    def visitDesign_statement(self, ctx:VerilogParser.Design_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#config_rule_statement.
    def visitConfig_rule_statement(self, ctx:VerilogParser.Config_rule_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#default_clause.
    def visitDefault_clause(self, ctx:VerilogParser.Default_clauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#inst_clause.
    def visitInst_clause(self, ctx:VerilogParser.Inst_clauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#inst_name.
    def visitInst_name(self, ctx:VerilogParser.Inst_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#cell_clause.
    def visitCell_clause(self, ctx:VerilogParser.Cell_clauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#liblist_clause.
    def visitLiblist_clause(self, ctx:VerilogParser.Liblist_clauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#use_clause.
    def visitUse_clause(self, ctx:VerilogParser.Use_clauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#local_parameter_declaration.
    def visitLocal_parameter_declaration(self, ctx:VerilogParser.Local_parameter_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#parameter_declaration.
    def visitParameter_declaration(self, ctx:VerilogParser.Parameter_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#specparam_declaration.
    def visitSpecparam_declaration(self, ctx:VerilogParser.Specparam_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#parameter_type.
    def visitParameter_type(self, ctx:VerilogParser.Parameter_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#inout_declaration.
    def visitInout_declaration(self, ctx:VerilogParser.Inout_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#input_declaration.
    def visitInput_declaration(self, ctx:VerilogParser.Input_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#output_declaration.
    def visitOutput_declaration(self, ctx:VerilogParser.Output_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#event_declaration.
    def visitEvent_declaration(self, ctx:VerilogParser.Event_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#integer_declaration.
    def visitInteger_declaration(self, ctx:VerilogParser.Integer_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#net_declaration.
    def visitNet_declaration(self, ctx:VerilogParser.Net_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#real_declaration.
    def visitReal_declaration(self, ctx:VerilogParser.Real_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#realtime_declaration.
    def visitRealtime_declaration(self, ctx:VerilogParser.Realtime_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#reg_declaration.
    def visitReg_declaration(self, ctx:VerilogParser.Reg_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#time_declaration.
    def visitTime_declaration(self, ctx:VerilogParser.Time_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#net_type.
    def visitNet_type(self, ctx:VerilogParser.Net_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#output_variable_type.
    def visitOutput_variable_type(self, ctx:VerilogParser.Output_variable_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#real_type.
    def visitReal_type(self, ctx:VerilogParser.Real_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#variable_type.
    def visitVariable_type(self, ctx:VerilogParser.Variable_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#drive_strength.
    def visitDrive_strength(self, ctx:VerilogParser.Drive_strengthContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#strength0.
    def visitStrength0(self, ctx:VerilogParser.Strength0Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#strength1.
    def visitStrength1(self, ctx:VerilogParser.Strength1Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#charge_strength.
    def visitCharge_strength(self, ctx:VerilogParser.Charge_strengthContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#delay3.
    def visitDelay3(self, ctx:VerilogParser.Delay3Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#delay2.
    def visitDelay2(self, ctx:VerilogParser.Delay2Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#delay_value.
    def visitDelay_value(self, ctx:VerilogParser.Delay_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_defparam_assignments.
    def visitList_of_defparam_assignments(self, ctx:VerilogParser.List_of_defparam_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_event_identifiers.
    def visitList_of_event_identifiers(self, ctx:VerilogParser.List_of_event_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_net_decl_assignments.
    def visitList_of_net_decl_assignments(self, ctx:VerilogParser.List_of_net_decl_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_net_identifiers.
    def visitList_of_net_identifiers(self, ctx:VerilogParser.List_of_net_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_param_assignments.
    def visitList_of_param_assignments(self, ctx:VerilogParser.List_of_param_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_port_identifiers.
    def visitList_of_port_identifiers(self, ctx:VerilogParser.List_of_port_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_real_identifiers.
    def visitList_of_real_identifiers(self, ctx:VerilogParser.List_of_real_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_specparam_assignments.
    def visitList_of_specparam_assignments(self, ctx:VerilogParser.List_of_specparam_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_variable_identifiers.
    def visitList_of_variable_identifiers(self, ctx:VerilogParser.List_of_variable_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_variable_port_identifiers.
    def visitList_of_variable_port_identifiers(self, ctx:VerilogParser.List_of_variable_port_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#defparam_assignment.
    def visitDefparam_assignment(self, ctx:VerilogParser.Defparam_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#net_decl_assignment.
    def visitNet_decl_assignment(self, ctx:VerilogParser.Net_decl_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#param_assignment.
    def visitParam_assignment(self, ctx:VerilogParser.Param_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#specparam_assignment.
    def visitSpecparam_assignment(self, ctx:VerilogParser.Specparam_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#pulse_control_specparam.
    def visitPulse_control_specparam(self, ctx:VerilogParser.Pulse_control_specparamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#error_limit_value.
    def visitError_limit_value(self, ctx:VerilogParser.Error_limit_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#reject_limit_value.
    def visitReject_limit_value(self, ctx:VerilogParser.Reject_limit_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#limit_value.
    def visitLimit_value(self, ctx:VerilogParser.Limit_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dimension.
    def visitDimension(self, ctx:VerilogParser.DimensionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#range_.
    def visitRange_(self, ctx:VerilogParser.Range_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#function_declaration.
    def visitFunction_declaration(self, ctx:VerilogParser.Function_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#function_item_declaration.
    def visitFunction_item_declaration(self, ctx:VerilogParser.Function_item_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#function_port_list.
    def visitFunction_port_list(self, ctx:VerilogParser.Function_port_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#function_range_or_type.
    def visitFunction_range_or_type(self, ctx:VerilogParser.Function_range_or_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#task_declaration.
    def visitTask_declaration(self, ctx:VerilogParser.Task_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#task_item_declaration.
    def visitTask_item_declaration(self, ctx:VerilogParser.Task_item_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#task_port_list.
    def visitTask_port_list(self, ctx:VerilogParser.Task_port_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#task_port_item.
    def visitTask_port_item(self, ctx:VerilogParser.Task_port_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#tf_input_declaration.
    def visitTf_input_declaration(self, ctx:VerilogParser.Tf_input_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#tf_output_declaration.
    def visitTf_output_declaration(self, ctx:VerilogParser.Tf_output_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#tf_inout_declaration.
    def visitTf_inout_declaration(self, ctx:VerilogParser.Tf_inout_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#task_port_type.
    def visitTask_port_type(self, ctx:VerilogParser.Task_port_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#block_item_declaration.
    def visitBlock_item_declaration(self, ctx:VerilogParser.Block_item_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_block_variable_identifiers.
    def visitList_of_block_variable_identifiers(self, ctx:VerilogParser.List_of_block_variable_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_block_real_identifiers.
    def visitList_of_block_real_identifiers(self, ctx:VerilogParser.List_of_block_real_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#block_variable_type.
    def visitBlock_variable_type(self, ctx:VerilogParser.Block_variable_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#block_real_type.
    def visitBlock_real_type(self, ctx:VerilogParser.Block_real_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#gate_instantiation.
    def visitGate_instantiation(self, ctx:VerilogParser.Gate_instantiationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#cmos_switch_instance.
    def visitCmos_switch_instance(self, ctx:VerilogParser.Cmos_switch_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#enable_gate_instance.
    def visitEnable_gate_instance(self, ctx:VerilogParser.Enable_gate_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#mos_switch_instance.
    def visitMos_switch_instance(self, ctx:VerilogParser.Mos_switch_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#n_input_gate_instance.
    def visitN_input_gate_instance(self, ctx:VerilogParser.N_input_gate_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#n_output_gate_instance.
    def visitN_output_gate_instance(self, ctx:VerilogParser.N_output_gate_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#pass_switch_instance.
    def visitPass_switch_instance(self, ctx:VerilogParser.Pass_switch_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#pass_enable_switch_instance.
    def visitPass_enable_switch_instance(self, ctx:VerilogParser.Pass_enable_switch_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#pull_gate_instance.
    def visitPull_gate_instance(self, ctx:VerilogParser.Pull_gate_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#name_of_gate_instance.
    def visitName_of_gate_instance(self, ctx:VerilogParser.Name_of_gate_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#pulldown_strength.
    def visitPulldown_strength(self, ctx:VerilogParser.Pulldown_strengthContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#pullup_strength.
    def visitPullup_strength(self, ctx:VerilogParser.Pullup_strengthContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#enable_terminal.
    def visitEnable_terminal(self, ctx:VerilogParser.Enable_terminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#inout_terminal.
    def visitInout_terminal(self, ctx:VerilogParser.Inout_terminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#input_terminal.
    def visitInput_terminal(self, ctx:VerilogParser.Input_terminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#ncontrol_terminal.
    def visitNcontrol_terminal(self, ctx:VerilogParser.Ncontrol_terminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#output_terminal.
    def visitOutput_terminal(self, ctx:VerilogParser.Output_terminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#pcontrol_terminal.
    def visitPcontrol_terminal(self, ctx:VerilogParser.Pcontrol_terminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#cmos_switchtype.
    def visitCmos_switchtype(self, ctx:VerilogParser.Cmos_switchtypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#enable_gatetype.
    def visitEnable_gatetype(self, ctx:VerilogParser.Enable_gatetypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#mos_switchtype.
    def visitMos_switchtype(self, ctx:VerilogParser.Mos_switchtypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#n_input_gatetype.
    def visitN_input_gatetype(self, ctx:VerilogParser.N_input_gatetypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#n_output_gatetype.
    def visitN_output_gatetype(self, ctx:VerilogParser.N_output_gatetypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#pass_en_switchtype.
    def visitPass_en_switchtype(self, ctx:VerilogParser.Pass_en_switchtypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#pass_switchtype.
    def visitPass_switchtype(self, ctx:VerilogParser.Pass_switchtypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_instantiation.
    def visitModule_instantiation(self, ctx:VerilogParser.Module_instantiationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#parameter_value_assignment.
    def visitParameter_value_assignment(self, ctx:VerilogParser.Parameter_value_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_parameter_assignments.
    def visitList_of_parameter_assignments(self, ctx:VerilogParser.List_of_parameter_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#ordered_parameter_assignment.
    def visitOrdered_parameter_assignment(self, ctx:VerilogParser.Ordered_parameter_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#named_parameter_assignment.
    def visitNamed_parameter_assignment(self, ctx:VerilogParser.Named_parameter_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_instance.
    def visitModule_instance(self, ctx:VerilogParser.Module_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#name_of_module_instance.
    def visitName_of_module_instance(self, ctx:VerilogParser.Name_of_module_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_port_connections.
    def visitList_of_port_connections(self, ctx:VerilogParser.List_of_port_connectionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#ordered_port_connection.
    def visitOrdered_port_connection(self, ctx:VerilogParser.Ordered_port_connectionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#named_port_connection.
    def visitNamed_port_connection(self, ctx:VerilogParser.Named_port_connectionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#generate_region.
    def visitGenerate_region(self, ctx:VerilogParser.Generate_regionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#genvar_declaration.
    def visitGenvar_declaration(self, ctx:VerilogParser.Genvar_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_genvar_identifiers.
    def visitList_of_genvar_identifiers(self, ctx:VerilogParser.List_of_genvar_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#loop_generate_construct.
    def visitLoop_generate_construct(self, ctx:VerilogParser.Loop_generate_constructContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#genvar_initialization.
    def visitGenvar_initialization(self, ctx:VerilogParser.Genvar_initializationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#genvar_expression.
    def visitGenvar_expression(self, ctx:VerilogParser.Genvar_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#genvar_iteration.
    def visitGenvar_iteration(self, ctx:VerilogParser.Genvar_iterationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#genvar_primary.
    def visitGenvar_primary(self, ctx:VerilogParser.Genvar_primaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#conditional_generate_construct.
    def visitConditional_generate_construct(self, ctx:VerilogParser.Conditional_generate_constructContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#if_generate_construct.
    def visitIf_generate_construct(self, ctx:VerilogParser.If_generate_constructContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#case_generate_construct.
    def visitCase_generate_construct(self, ctx:VerilogParser.Case_generate_constructContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#generate_block.
    def visitGenerate_block(self, ctx:VerilogParser.Generate_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#generate_block_or_null.
    def visitGenerate_block_or_null(self, ctx:VerilogParser.Generate_block_or_nullContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#continuous_assign.
    def visitContinuous_assign(self, ctx:VerilogParser.Continuous_assignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_net_assignments.
    def visitList_of_net_assignments(self, ctx:VerilogParser.List_of_net_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#net_assignment.
    def visitNet_assignment(self, ctx:VerilogParser.Net_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#initial_construct.
    def visitInitial_construct(self, ctx:VerilogParser.Initial_constructContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#always_construct.
    def visitAlways_construct(self, ctx:VerilogParser.Always_constructContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#blocking_assignment.
    def visitBlocking_assignment(self, ctx:VerilogParser.Blocking_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#nonblocking_assignment.
    def visitNonblocking_assignment(self, ctx:VerilogParser.Nonblocking_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#procedural_continuous_assignments.
    def visitProcedural_continuous_assignments(self, ctx:VerilogParser.Procedural_continuous_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#variable_assignment.
    def visitVariable_assignment(self, ctx:VerilogParser.Variable_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#par_block.
    def visitPar_block(self, ctx:VerilogParser.Par_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#seq_block.
    def visitSeq_block(self, ctx:VerilogParser.Seq_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#statement.
    def visitStatement(self, ctx:VerilogParser.StatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#statement_or_null.
    def visitStatement_or_null(self, ctx:VerilogParser.Statement_or_nullContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#function_statement.
    def visitFunction_statement(self, ctx:VerilogParser.Function_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#delay_control.
    def visitDelay_control(self, ctx:VerilogParser.Delay_controlContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#delay_or_event_control.
    def visitDelay_or_event_control(self, ctx:VerilogParser.Delay_or_event_controlContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#disable_statement.
    def visitDisable_statement(self, ctx:VerilogParser.Disable_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#event_control.
    def visitEvent_control(self, ctx:VerilogParser.Event_controlContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#event_trigger.
    def visitEvent_trigger(self, ctx:VerilogParser.Event_triggerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#event_expression.
    def visitEvent_expression(self, ctx:VerilogParser.Event_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#event_primary.
    def visitEvent_primary(self, ctx:VerilogParser.Event_primaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#procedural_timing_control.
    def visitProcedural_timing_control(self, ctx:VerilogParser.Procedural_timing_controlContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#procedural_timing_control_statement.
    def visitProcedural_timing_control_statement(self, ctx:VerilogParser.Procedural_timing_control_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#wait_statement.
    def visitWait_statement(self, ctx:VerilogParser.Wait_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#conditional_statement.
    def visitConditional_statement(self, ctx:VerilogParser.Conditional_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#case_statement.
    def visitCase_statement(self, ctx:VerilogParser.Case_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#case_item.
    def visitCase_item(self, ctx:VerilogParser.Case_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#loop_statement.
    def visitLoop_statement(self, ctx:VerilogParser.Loop_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#system_task_enable.
    def visitSystem_task_enable(self, ctx:VerilogParser.System_task_enableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#task_enable.
    def visitTask_enable(self, ctx:VerilogParser.Task_enableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#specify_block.
    def visitSpecify_block(self, ctx:VerilogParser.Specify_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#specify_item.
    def visitSpecify_item(self, ctx:VerilogParser.Specify_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#pulsestyle_declaration.
    def visitPulsestyle_declaration(self, ctx:VerilogParser.Pulsestyle_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#showcancelled_declaration.
    def visitShowcancelled_declaration(self, ctx:VerilogParser.Showcancelled_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#path_declaration.
    def visitPath_declaration(self, ctx:VerilogParser.Path_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#simple_path_declaration.
    def visitSimple_path_declaration(self, ctx:VerilogParser.Simple_path_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#parallel_path_description.
    def visitParallel_path_description(self, ctx:VerilogParser.Parallel_path_descriptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#full_path_description.
    def visitFull_path_description(self, ctx:VerilogParser.Full_path_descriptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_path_inputs.
    def visitList_of_path_inputs(self, ctx:VerilogParser.List_of_path_inputsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_path_outputs.
    def visitList_of_path_outputs(self, ctx:VerilogParser.List_of_path_outputsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#specify_input_terminal_descriptor.
    def visitSpecify_input_terminal_descriptor(self, ctx:VerilogParser.Specify_input_terminal_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#specify_output_terminal_descriptor.
    def visitSpecify_output_terminal_descriptor(self, ctx:VerilogParser.Specify_output_terminal_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#input_identifier.
    def visitInput_identifier(self, ctx:VerilogParser.Input_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#output_identifier.
    def visitOutput_identifier(self, ctx:VerilogParser.Output_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#path_delay_value.
    def visitPath_delay_value(self, ctx:VerilogParser.Path_delay_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#list_of_path_delay_expressions.
    def visitList_of_path_delay_expressions(self, ctx:VerilogParser.List_of_path_delay_expressionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#t_path_delay_expression.
    def visitT_path_delay_expression(self, ctx:VerilogParser.T_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#trise_path_delay_expression.
    def visitTrise_path_delay_expression(self, ctx:VerilogParser.Trise_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#tfall_path_delay_expression.
    def visitTfall_path_delay_expression(self, ctx:VerilogParser.Tfall_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#tz_path_delay_expression.
    def visitTz_path_delay_expression(self, ctx:VerilogParser.Tz_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#t01_path_delay_expression.
    def visitT01_path_delay_expression(self, ctx:VerilogParser.T01_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#t10_path_delay_expression.
    def visitT10_path_delay_expression(self, ctx:VerilogParser.T10_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#t0z_path_delay_expression.
    def visitT0z_path_delay_expression(self, ctx:VerilogParser.T0z_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#tz1_path_delay_expression.
    def visitTz1_path_delay_expression(self, ctx:VerilogParser.Tz1_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#t1z_path_delay_expression.
    def visitT1z_path_delay_expression(self, ctx:VerilogParser.T1z_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#tz0_path_delay_expression.
    def visitTz0_path_delay_expression(self, ctx:VerilogParser.Tz0_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#t0x_path_delay_expression.
    def visitT0x_path_delay_expression(self, ctx:VerilogParser.T0x_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#tx1_path_delay_expression.
    def visitTx1_path_delay_expression(self, ctx:VerilogParser.Tx1_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#t1x_path_delay_expression.
    def visitT1x_path_delay_expression(self, ctx:VerilogParser.T1x_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#tx0_path_delay_expression.
    def visitTx0_path_delay_expression(self, ctx:VerilogParser.Tx0_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#txz_path_delay_expression.
    def visitTxz_path_delay_expression(self, ctx:VerilogParser.Txz_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#tzx_path_delay_expression.
    def visitTzx_path_delay_expression(self, ctx:VerilogParser.Tzx_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#path_delay_expression.
    def visitPath_delay_expression(self, ctx:VerilogParser.Path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#edge_sensitive_path_declaration.
    def visitEdge_sensitive_path_declaration(self, ctx:VerilogParser.Edge_sensitive_path_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#parallel_edge_sensitive_path_description.
    def visitParallel_edge_sensitive_path_description(self, ctx:VerilogParser.Parallel_edge_sensitive_path_descriptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#full_edge_sensitive_path_description.
    def visitFull_edge_sensitive_path_description(self, ctx:VerilogParser.Full_edge_sensitive_path_descriptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#data_source_expression.
    def visitData_source_expression(self, ctx:VerilogParser.Data_source_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#edge_identifier.
    def visitEdge_identifier(self, ctx:VerilogParser.Edge_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#state_dependent_path_declaration.
    def visitState_dependent_path_declaration(self, ctx:VerilogParser.State_dependent_path_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#polarity_operator.
    def visitPolarity_operator(self, ctx:VerilogParser.Polarity_operatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#concatenation.
    def visitConcatenation(self, ctx:VerilogParser.ConcatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#constant_concatenation.
    def visitConstant_concatenation(self, ctx:VerilogParser.Constant_concatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#constant_multiple_concatenation.
    def visitConstant_multiple_concatenation(self, ctx:VerilogParser.Constant_multiple_concatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_path_concatenation.
    def visitModule_path_concatenation(self, ctx:VerilogParser.Module_path_concatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_path_multiple_concatenation.
    def visitModule_path_multiple_concatenation(self, ctx:VerilogParser.Module_path_multiple_concatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#multiple_concatenation.
    def visitMultiple_concatenation(self, ctx:VerilogParser.Multiple_concatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#constant_function_call.
    def visitConstant_function_call(self, ctx:VerilogParser.Constant_function_callContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#constant_system_function_call.
    def visitConstant_system_function_call(self, ctx:VerilogParser.Constant_system_function_callContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#function_call.
    def visitFunction_call(self, ctx:VerilogParser.Function_callContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#system_function_call.
    def visitSystem_function_call(self, ctx:VerilogParser.System_function_callContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#base_expression.
    def visitBase_expression(self, ctx:VerilogParser.Base_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#constant_base_expression.
    def visitConstant_base_expression(self, ctx:VerilogParser.Constant_base_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#constant_expression.
    def visitConstant_expression(self, ctx:VerilogParser.Constant_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#constant_mintypmax_expression.
    def visitConstant_mintypmax_expression(self, ctx:VerilogParser.Constant_mintypmax_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#constant_range_expression.
    def visitConstant_range_expression(self, ctx:VerilogParser.Constant_range_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#dimension_constant_expression.
    def visitDimension_constant_expression(self, ctx:VerilogParser.Dimension_constant_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#expression.
    def visitExpression(self, ctx:VerilogParser.ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#lsb_constant_expression.
    def visitLsb_constant_expression(self, ctx:VerilogParser.Lsb_constant_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#mintypmax_expression.
    def visitMintypmax_expression(self, ctx:VerilogParser.Mintypmax_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_path_expression.
    def visitModule_path_expression(self, ctx:VerilogParser.Module_path_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_path_mintypmax_expression.
    def visitModule_path_mintypmax_expression(self, ctx:VerilogParser.Module_path_mintypmax_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#msb_constant_expression.
    def visitMsb_constant_expression(self, ctx:VerilogParser.Msb_constant_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#range_expression.
    def visitRange_expression(self, ctx:VerilogParser.Range_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#width_constant_expression.
    def visitWidth_constant_expression(self, ctx:VerilogParser.Width_constant_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#constant_primary.
    def visitConstant_primary(self, ctx:VerilogParser.Constant_primaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_path_primary.
    def visitModule_path_primary(self, ctx:VerilogParser.Module_path_primaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#primary.
    def visitPrimary(self, ctx:VerilogParser.PrimaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#net_lvalue.
    def visitNet_lvalue(self, ctx:VerilogParser.Net_lvalueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#variable_lvalue.
    def visitVariable_lvalue(self, ctx:VerilogParser.Variable_lvalueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#unary_operator.
    def visitUnary_operator(self, ctx:VerilogParser.Unary_operatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#binary_operator.
    def visitBinary_operator(self, ctx:VerilogParser.Binary_operatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#unary_module_path_operator.
    def visitUnary_module_path_operator(self, ctx:VerilogParser.Unary_module_path_operatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#binary_module_path_operator.
    def visitBinary_module_path_operator(self, ctx:VerilogParser.Binary_module_path_operatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#number.
    def visitNumber(self, ctx:VerilogParser.NumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#attribute_instance.
    def visitAttribute_instance(self, ctx:VerilogParser.Attribute_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#attr_spec.
    def visitAttr_spec(self, ctx:VerilogParser.Attr_specContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#attr_name.
    def visitAttr_name(self, ctx:VerilogParser.Attr_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#block_identifier.
    def visitBlock_identifier(self, ctx:VerilogParser.Block_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#cell_identifier.
    def visitCell_identifier(self, ctx:VerilogParser.Cell_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#config_identifier.
    def visitConfig_identifier(self, ctx:VerilogParser.Config_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#event_identifier.
    def visitEvent_identifier(self, ctx:VerilogParser.Event_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#function_identifier.
    def visitFunction_identifier(self, ctx:VerilogParser.Function_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#gate_instance_identifier.
    def visitGate_instance_identifier(self, ctx:VerilogParser.Gate_instance_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#generate_block_identifier.
    def visitGenerate_block_identifier(self, ctx:VerilogParser.Generate_block_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#genvar_identifier.
    def visitGenvar_identifier(self, ctx:VerilogParser.Genvar_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#hierarchical_block_identifier.
    def visitHierarchical_block_identifier(self, ctx:VerilogParser.Hierarchical_block_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#hierarchical_event_identifier.
    def visitHierarchical_event_identifier(self, ctx:VerilogParser.Hierarchical_event_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#hierarchical_function_identifier.
    def visitHierarchical_function_identifier(self, ctx:VerilogParser.Hierarchical_function_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#hierarchical_identifier.
    def visitHierarchical_identifier(self, ctx:VerilogParser.Hierarchical_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#hierarchical_net_identifier.
    def visitHierarchical_net_identifier(self, ctx:VerilogParser.Hierarchical_net_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#hierarchical_parameter_identifier.
    def visitHierarchical_parameter_identifier(self, ctx:VerilogParser.Hierarchical_parameter_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#hierarchical_variable_identifier.
    def visitHierarchical_variable_identifier(self, ctx:VerilogParser.Hierarchical_variable_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#hierarchical_task_identifier.
    def visitHierarchical_task_identifier(self, ctx:VerilogParser.Hierarchical_task_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#identifier.
    def visitIdentifier(self, ctx:VerilogParser.IdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#inout_port_identifier.
    def visitInout_port_identifier(self, ctx:VerilogParser.Inout_port_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#input_port_identifier.
    def visitInput_port_identifier(self, ctx:VerilogParser.Input_port_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#instance_identifier.
    def visitInstance_identifier(self, ctx:VerilogParser.Instance_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#library_identifier.
    def visitLibrary_identifier(self, ctx:VerilogParser.Library_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_identifier.
    def visitModule_identifier(self, ctx:VerilogParser.Module_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#module_instance_identifier.
    def visitModule_instance_identifier(self, ctx:VerilogParser.Module_instance_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#net_identifier.
    def visitNet_identifier(self, ctx:VerilogParser.Net_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#output_port_identifier.
    def visitOutput_port_identifier(self, ctx:VerilogParser.Output_port_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#parameter_identifier.
    def visitParameter_identifier(self, ctx:VerilogParser.Parameter_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#port_identifier.
    def visitPort_identifier(self, ctx:VerilogParser.Port_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#real_identifier.
    def visitReal_identifier(self, ctx:VerilogParser.Real_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#specparam_identifier.
    def visitSpecparam_identifier(self, ctx:VerilogParser.Specparam_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#system_function_identifier.
    def visitSystem_function_identifier(self, ctx:VerilogParser.System_function_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#system_task_identifier.
    def visitSystem_task_identifier(self, ctx:VerilogParser.System_task_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#task_identifier.
    def visitTask_identifier(self, ctx:VerilogParser.Task_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#terminal_identifier.
    def visitTerminal_identifier(self, ctx:VerilogParser.Terminal_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#topmodule_identifier.
    def visitTopmodule_identifier(self, ctx:VerilogParser.Topmodule_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#udp_identifier.
    def visitUdp_identifier(self, ctx:VerilogParser.Udp_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#udp_instance_identifier.
    def visitUdp_instance_identifier(self, ctx:VerilogParser.Udp_instance_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogParser#variable_identifier.
    def visitVariable_identifier(self, ctx:VerilogParser.Variable_identifierContext):
        return self.visitChildren(ctx)



del VerilogParser


# Generated from VerilogPreprocessorParser.g4 by ANTLR 4.9.3
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .VerilogPreprocessorParser import VerilogPreprocessorParser
else:
    from VerilogPreprocessorParser import VerilogPreprocessorParser

# This class defines a complete generic visitor for a parse tree produced by VerilogPreprocessorParser.

class VerilogPreprocessorParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by VerilogPreprocessorParser#source_text.
    def visitSource_text(self, ctx:VerilogPreprocessorParser.Source_textContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#compiler_directive.
    def visitCompiler_directive(self, ctx:VerilogPreprocessorParser.Compiler_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#celldefine_compiler_directive.
    def visitCelldefine_compiler_directive(self, ctx:VerilogPreprocessorParser.Celldefine_compiler_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#endcelldefine_compiler_directive.
    def visitEndcelldefine_compiler_directive(self, ctx:VerilogPreprocessorParser.Endcelldefine_compiler_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#default_nettype_compiler_directive.
    def visitDefault_nettype_compiler_directive(self, ctx:VerilogPreprocessorParser.Default_nettype_compiler_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#default_nettype_value.
    def visitDefault_nettype_value(self, ctx:VerilogPreprocessorParser.Default_nettype_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#text_macro_definition.
    def visitText_macro_definition(self, ctx:VerilogPreprocessorParser.Text_macro_definitionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#text_macro_usage.
    def visitText_macro_usage(self, ctx:VerilogPreprocessorParser.Text_macro_usageContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#macro_list_of_actual_arguments.
    def visitMacro_list_of_actual_arguments(self, ctx:VerilogPreprocessorParser.Macro_list_of_actual_argumentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#text_macro_identifier.
    def visitText_macro_identifier(self, ctx:VerilogPreprocessorParser.Text_macro_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#undefine_compiler_directive.
    def visitUndefine_compiler_directive(self, ctx:VerilogPreprocessorParser.Undefine_compiler_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#ifdef_directive.
    def visitIfdef_directive(self, ctx:VerilogPreprocessorParser.Ifdef_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#ifndef_directive.
    def visitIfndef_directive(self, ctx:VerilogPreprocessorParser.Ifndef_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#elsif_directive.
    def visitElsif_directive(self, ctx:VerilogPreprocessorParser.Elsif_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#else_directive.
    def visitElse_directive(self, ctx:VerilogPreprocessorParser.Else_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#endif_directive.
    def visitEndif_directive(self, ctx:VerilogPreprocessorParser.Endif_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#include_compiler_directive.
    def visitInclude_compiler_directive(self, ctx:VerilogPreprocessorParser.Include_compiler_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#filename.
    def visitFilename(self, ctx:VerilogPreprocessorParser.FilenameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#resetall_compiler_directive.
    def visitResetall_compiler_directive(self, ctx:VerilogPreprocessorParser.Resetall_compiler_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#line_compiler_directive.
    def visitLine_compiler_directive(self, ctx:VerilogPreprocessorParser.Line_compiler_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#line_number.
    def visitLine_number(self, ctx:VerilogPreprocessorParser.Line_numberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#line_level.
    def visitLine_level(self, ctx:VerilogPreprocessorParser.Line_levelContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#timescale_compiler_directive.
    def visitTimescale_compiler_directive(self, ctx:VerilogPreprocessorParser.Timescale_compiler_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#time_literal.
    def visitTime_literal(self, ctx:VerilogPreprocessorParser.Time_literalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#time_number.
    def visitTime_number(self, ctx:VerilogPreprocessorParser.Time_numberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#time_unit.
    def visitTime_unit(self, ctx:VerilogPreprocessorParser.Time_unitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#unconnected_drive_compiler_directive.
    def visitUnconnected_drive_compiler_directive(self, ctx:VerilogPreprocessorParser.Unconnected_drive_compiler_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#unconnected_drive_value.
    def visitUnconnected_drive_value(self, ctx:VerilogPreprocessorParser.Unconnected_drive_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#nounconnected_drive_compiler_directive.
    def visitNounconnected_drive_compiler_directive(self, ctx:VerilogPreprocessorParser.Nounconnected_drive_compiler_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#keywords_directive.
    def visitKeywords_directive(self, ctx:VerilogPreprocessorParser.Keywords_directiveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#version_specifier.
    def visitVersion_specifier(self, ctx:VerilogPreprocessorParser.Version_specifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by VerilogPreprocessorParser#endkeywords_directive.
    def visitEndkeywords_directive(self, ctx:VerilogPreprocessorParser.Endkeywords_directiveContext):
        return self.visitChildren(ctx)



del VerilogPreprocessorParser